@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000 Qualcomm Incorporated.  All rights reserved.
+ * Copyright (c) 2005 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  * 
@@ -18,6 +18,23 @@
  * 
  * Revisions:
  *
+ *   04/22/05 [rcg]
+ *        - Added code to use mkstemp() instead of our perfectly safe usage
+ *          of tempnam() because some compilers issue overly broad warnings
+ *          implying that all uses of tempnam() are unsafe.
+ *
+ *   04/06/01 [rcg]
+ *        - Fixed case where lock is usurped in error.  Thanks to Michael
+ *          Smith for finding this.
+ *
+ *   01/30/01 [rcg]
+ *        - Qmaillock now takes bNo_atomic_open and pSpool_dir parameters
+ *          instead of using compile-time constants.
+ *
+ *   01/19/01 [rcg]
+ *        - Modified to pass drop name instead of user name (based on
+ *          patch by Fergal Daly).
+ *
  *   07/17/00 [rcg]
  *        - Allow lock to proceed even if user is over quota.
  *
@@ -103,6 +120,7 @@
 #include "logit.h"
 #include "snprintf.h"
 #include "popper.h"
+#include "string_util.h"
 
 extern int gethostname();
 
@@ -165,6 +183,103 @@ touch_file ( char *filenm, void *fTrace )
 }
 
 
+/*
+ * Function:    temp_file
+ *
+ * Description: Opens a unique file.
+ *
+ * Parameters:  pDir:           Directory where file should be located.
+ *              pPfx:           Prefix for filename.
+ *              pTmp:           Address of pointer for constructed name.
+ *                              >> BE SURE TO FREE THIS WHEN YOU ARE DONE <<
+ *              fTrace:         FILE * for trace info, or NULL to use syslog.
+ *              bDebugging:     TRUE if debugging is desired.
+ *
+ * Returns:     file descriptor or -1.
+ */
+int
+temp_file ( const char *pDir, const char *pPfx, char **pTmp,
+            void *fTrace, int bDebugging )
+{
+    int   fd   =   -1;
+    int   rslt =   -1;
+    char *pStr = NULL;
+
+
+    /*
+     * Some compilers issue overly broad warnings which imply that
+     * all uses of tempnam() are unsafe.  This is false.  tempnam()
+     * is safe as long as the created string is tested by opening
+     * exclusive, which Qpopper has always done.  But, because people
+     * don't understand this and hence complain about the warning, the
+     * following code uses mkstemp() instead.  *sigh*
+     */
+
+#ifndef USE_TEMPNAME
+
+    pStr = malloc ( MAXDROPLEN );
+    if ( pStr == NULL )
+    {
+        *pTmp = NULL;
+        return ( -1 );
+    }
+    rslt = Qsnprintf ( pStr, MAXDROPLEN, "%s/%sXXXX", pDir, pPfx );
+    if ( rslt == -1 )
+    {
+        if ( DEBUGGING && bDebugging )
+            logit ( fTrace, POP_DEBUG, HERE,
+                    "Error creating temp file name template using "
+                    "dir %s and prefix %s",
+                    pDir, pPfx );
+        return ( -1 );
+    }
+    if ( DEBUGGING && bDebugging )
+        logit ( fTrace, POP_DEBUG, HERE,
+                "Created temp file name template: %s", pStr );
+
+    fd = mkstemp ( pStr );
+
+    if ( DEBUGGING && bDebugging )
+    {
+        if ( fd == -1 )
+            logit ( fTrace, POP_DEBUG, HERE,
+                    "mkstemp error: %s (%d)", strerror(errno), errno );
+        else
+            logit ( fTrace, POP_DEBUG, HERE,
+                    "mkstemp() generated: %s", pStr );
+    }
+
+    *pTmp = pStr;
+    return ( fd );
+
+#else
+
+    pStr = tempnam ( pDir, pPfx );
+    if ( pStr == NULL )
+        return -1;
+    fd = open ( pStr, O_CREAT | O_EXCL, 0600 );
+    if ( fd == -1 )
+    {
+        if ( DEBUGGING && bDebugging )
+        {
+            logit ( fTrace, POP_DEBUG, HERE,
+                    "open() for temp file %s returned %s (%d)",
+                    pTmp, strerror(errno), errno );
+                if ( errno == EEXIST )
+                    logit ( fTrace, POP_DEBUG, HERE,
+                            "temp file %s exists", pTmp )
+                else
+                    logit ( fTrace, POP_PRIORITY, HERE,
+                            "Unable to create temp file (for lock) %s: %s (%d)",
+                            pTmp, strerror(errno), errno );
+        }
+    }
+    return ( fd );
+
+#endif /* USE_TEMPNAME */
+}
+
+
 /*
  * Function:    maillock
  *
@@ -173,13 +288,15 @@ touch_file ( char *filenm, void *fTrace )
  *              Every maillock should be paired with mailunlock before a
  *              subsequent call to maillock() is made.
  *
- * Parameters:  usernm:     User id.
- *              retrycnt:   Retry Count. Between retries there is a sleep of 5
- *                          seconds times the retrycount.
- *              fTrace:     file ID for trace info, or NULL to use syslog.
- *              fn:         file name whence called.
- *              ln:         line number whence called.
- *              bDebugging: TRUE if debugging is desired.
+ * Parameters:  drop_name:       Name of file to be locked (spool name).
+ *              retrycnt:        Retry Count. Between retries there is a
+ *                                  sleep of 5 seconds times the retrycount.
+ *              bNo_atomic_open: TRUE if open() isn't automic.
+ *              pSpool_dir:      Spool directory.
+ *              fTrace:          FILE * for trace info, or NULL to use syslog.
+ *              fn:              File name whence called.
+ *              ln:              Line number whence called.
+ *              bDebugging:      TRUE if debugging is desired.
  *
  * Returns:     L_SUCCESS on success. 
  *
@@ -187,8 +304,10 @@ touch_file ( char *filenm, void *fTrace )
  */
 
 int 
-Qmaillock ( char         *usernm,
+Qmaillock ( char         *drop_name,
             int           retrycnt,
+            BOOL          bNo_atomic_open,
+            char         *pSpool_dir,
             void         *fTrace,
             const char   *fn,
             size_t        ln,
@@ -198,11 +317,11 @@ Qmaillock ( char         *usernm,
     struct stat   byNameStat;
     int           attemptcnt   = 0;
     time_t        curTime;
-    int           nRet         = L_SUCCESS;
-
-#if defined(NO_ATOMIC_OPEN)
+    int           nRet         = L_ERROR;
+    int           nBufSize     = sizeof(szLock);
+    int           nLen1        = 0;
+    int           nLen2        = 0;
     char         *pTmp         = NULL;
-#endif
 
     /*
      * Save trace file pointer and bDebugging flag for use in 
@@ -212,17 +331,20 @@ Qmaillock ( char         *usernm,
     gbDebugging = bDebugging;
 
     /*
-     * Get the path to the user's mail drop.
-     * Append ".lock" to name.
+     * Append ".lock" to spool name.
      */
-   
-    if ( genpath ( usernm, szLock, sizeof(szLock), 
-                   GNPH_SPOOL, fTrace, bDebugging ) < 0 ) {
+    nLen1 = strlcpy ( szLock, drop_name, nBufSize );
+    nLen2 = strlcat ( szLock, ".lock",   nBufSize );
+
+    /*
+     * See if we ran out of room with the drop_name, ".lock" and a NULL.
+     */
+    if ( nLen1 > nBufSize || nLen2 > nBufSize )
+    {
         logit ( fTrace, POP_PRIORITY, HERE,
-                "genpath() returned error for user %s", usernm );
+                "insufficient room to generate lock file for %s", drop_name );
         return L_ERROR;
     }
-    strcat ( szLock, ".lock" );
 
     /*
      * Verify if lockfile is a symbolic link, or a hard link.
@@ -246,192 +368,202 @@ Qmaillock ( char         *usernm,
         return L_ERROR;
     }
 
-    while ( attemptcnt != retrycnt ) {
+    while ( nRet != L_SUCCESS && attemptcnt < retrycnt ) {
         if ( attemptcnt != 0 ) 
             sleep ( attemptcnt * SLEEP_INTERVAL );  /* Sleep between retries */
 
-#if defined(NO_ATOMIC_OPEN)
-        while ( fd == -1 ) {
-            /* 
-             * Create a temporary file and link it to lock file 
-             */
-            pTmp = tempnam ( POP_MAILDIR, "POP" );
-            if ( pTmp != NULL )
-                fd = open ( pTmp, O_CREAT | O_EXCL, 0600 );
-            if ( fd == -1 ) {
-                if ( errno == EEXIST ) {
+        if ( bNo_atomic_open ) {
+            while ( fd == -1 ) {
+                /* 
+                 * Create a temporary file and link it to lock file 
+                 */
+                fd = temp_file ( pSpool_dir, "POP", &pTmp, fTrace, bDebugging );
+                if ( fd == -1 ) {
                     if ( pTmp != NULL )
                         free ( pTmp );
                     pTmp = NULL;
-                    continue;
-                    }
-                else {
-                    logit ( fTrace, POP_PRIORITY, HERE,
-                            "Unable to create temp file (for lock) %s: %s (%d)",
-                            pTmp, strerror(errno), errno );
-                    nRet = L_ERROR;
-                    attemptcnt = retrycnt;
-                    if ( pTmp != NULL )
-                        free ( pTmp );
-                    pTmp = NULL;
-                    break;
-                } /* unexpected error */
-            } /* open failed */
-        } /* while loop */
+                    if ( errno == EEXIST )
+                        continue;
+                    else {
+                        nRet = L_ERROR;
+                        break;
+                    } /* unexpected error */
+                } /* open failed */
+            } /* while loop */
 
-        if ( pTmp != NULL && link ( pTmp, szLock ) == -1 ) {
-            if ( errno == EEXIST ) {
-                curTime = time(0);
-                if ( stat ( szLock, &byNameStat ) != -1 && 
-                     curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
-                    unlink ( szLock );                   /* Remove the old lock */
-                    if ( link ( pTmp, szLock ) == -1 ) { /* Place a new lock */
-                        nRet = L_TMPLOCK;
-                        logit ( fTrace, POP_PRIORITY, HERE,
-                                "Unable to create link from temp file %s to new "
+            if ( pTmp != NULL && link ( pTmp, szLock ) == -1 ) {
+                if ( errno == EEXIST ) {
+                    curTime = time(0);
+                    if ( stat ( szLock, &byNameStat ) != -1 && 
+                         curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
+                        unlink ( szLock );                   /* Remove the old lock */
+                        if ( link ( pTmp, szLock ) == -1 ) { /* Place a new lock */
+                            nRet = L_TMPLOCK;
+                            logit ( fTrace, POP_PRIORITY, HERE,
+                                    "Unable to create link from temp file %s to "
+                                    "new lock file %s: %s (%d)",
+                                    pTmp, szLock, strerror(errno), errno );
+                        } /* link failed */
+                        attemptcnt = retrycnt;              /* Break out */
+                    } /* lock present and unexpired */
+                    else
+                        attemptcnt++;                       /* Try again */
+                } /* EEXIST error */
+                else {
+                    if ( DEBUGGING && bDebugging ) {
+                        logit ( fTrace, POP_DEBUG, HERE,
+                                "Unable to create link from temp file %s to "
                                 "lock file %s: %s (%d)",
                                 pTmp, szLock, strerror(errno), errno );
-                    } /* link failed */
-                    attemptcnt = retrycnt;              /* Break out */
-                } /* lock present and unexpired */
-                else
-                    attemptcnt++;                       /* Try again */
-            } /* EEXIST error */
+                    }
+                    nRet =  L_ERROR;
+                    attemptcnt = retrycnt;
+                } /* unexpected error */
+            } /* link or tempnam failed */
             else {
                 if ( DEBUGGING && bDebugging ) {
                     logit ( fTrace, POP_DEBUG, HERE,
-                            "Unable to create link from temp file %s to lock file %s: %s (%d)",
-                            pTmp, szLock, strerror(errno), errno );
+                            "successfully linked temp file %s to lock file %s",
+                            pTmp, szLock );
                 }
-                nRet =  L_ERROR;
-                attemptcnt = retrycnt;
-            } /* unexpected error */
-        } /* link or tempnam failed */
-        else {
-            attemptcnt = retrycnt;
+                nRet = L_SUCCESS;
+                if ( pTmp != NULL ) {
+                    unlink ( pTmp );
+                    free   ( pTmp );
+                    pTmp = NULL;
+                }
+            }
+
+            if ( nRet != L_SUCCESS && fd != -1 ) {
+                if ( pTmp != NULL )
+                    unlink ( pTmp );
+            }
+
             if ( pTmp != NULL ) {
-                unlink ( pTmp );
-                free   ( pTmp );
+                free ( pTmp );
                 pTmp = NULL;
             }
-        }
-        if ( nRet != L_SUCCESS && fd != -1 ) {
-            if ( pTmp != NULL )
-                unlink ( pTmp );
-        }
-        if ( pTmp != NULL ) {
-            free ( pTmp );
-            pTmp = NULL;
-        }
-        fd = -1;
-#else /* if defined(NO_ATOMIC_OPEN) */
- 
-        /* 
-         * The following is supposed to be atomic -- even on NFS.
-         */
-        fd = open ( szLock, O_WRONLY|O_CREAT|O_EXCL, 0600 );
-        if ( fd == -1 ) {
-            if ( DEBUGGING && bDebugging ) {
-                logit ( fTrace, POP_DEBUG, HERE,
-                        "open() for %s returned %s (%d)",
-                        szLock, strerror(errno), errno );
-            }
-            
-            switch ( errno ) {
-
-            case EACCES:
-                /* 
-                 * Keep trying for the lock life 
-                 */
-                curTime = time(0);
-                nRet = stat ( szLock, &byNameStat );
-                if ( nRet == -1 ) {
-                   logit ( fTrace, POP_PRIORITY, HERE,
-                           "fstat() for %s returned %s (%d) and open() returned EACCESS",
-                           szLock, strerror(errno), errno );
-                   attemptcnt = retrycnt;
-                   nRet = L_ERROR;
-                } 
-                else if ( curTime - (time_t)(byNameStat.st_mtime) < LOCK_LIFE ) {
-                    attemptcnt++;
-                    if ( DEBUGGING && bDebugging ) {
-                        logit ( fTrace, POP_DEBUG, HERE, 
-                                "lock %s busy", szLock );
-                    }
-                } 
-                else {
-                    attemptcnt = retrycnt;
-                    nRet = L_ERROR;
+            fd = -1;
+        } /* bNo_atomic_open == TRUE */
+        else { 
+            /* 
+             * The following is supposed to be atomic -- even on NFS (but
+             * we don't believe it).
+             */
+            fd = open ( szLock, O_WRONLY | O_CREAT | O_EXCL, 0600 );
+            if ( fd == -1 ) {
+                if ( DEBUGGING && bDebugging ) {
                     logit ( fTrace, POP_DEBUG, HERE,
-                            "Lock %s expired but access denied (%lu < %u)",
-                            szLock, curTime - (time_t)(byNameStat.st_mtime), 
-                            LOCK_LIFE );
+                            "open() for %s returned %s (%d)",
+                            szLock, strerror(errno), errno );
                 }
-                break;
+                
+                switch ( errno ) {
 
-            case EEXIST:
-                /* 
-                 * Check the lock life 
-                 */
-                curTime = time(0);
-                nRet = stat ( szLock, &byNameStat );
-                if ( nRet == -1 ) {
-                    logit ( fTrace, POP_PRIORITY, HERE,
-                           "fstat() for %s returned %s (%d) and "
-                           "open() returned EEXIST",
-                           szLock, strerror(errno), errno );
-                    attemptcnt = retrycnt;
-                    nRet = L_ERROR;
-                } 
-                else if ( curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
+                case EACCES:
                     /* 
-                     * Create the lock file.
-                     *
-                     * This create could be happening to a link. 
-                     * Check the race 
+                     * Keep trying for the lock life 
                      */
-                    fd = open ( szLock, O_WRONLY|O_CREAT|O_TRUNC, 0600 );
-                    if ( fd == -1 ) {
+                    curTime = time(0);
+                    nRet = stat ( szLock, &byNameStat );
+                    if ( nRet == -1 ) {
+                       logit ( fTrace, POP_PRIORITY, HERE,
+                               "fstat() for %s returned %s (%d) and "
+                               "open() returned EACCESS",
+                               szLock, strerror(errno), errno );
+                       attemptcnt = retrycnt;
+                       nRet = L_ERROR;
+                    } 
+                    else if ( curTime - (time_t)(byNameStat.st_mtime) < LOCK_LIFE ) {
+                        attemptcnt++;
+                        if ( DEBUGGING && bDebugging ) {
+                            logit ( fTrace, POP_DEBUG, HERE, 
+                                    "lock %s busy", szLock );
+                        }
+                    } 
+                    else {
+                        attemptcnt = retrycnt;
+                        nRet = L_ERROR;
+                        logit ( fTrace, POP_DEBUG, HERE,
+                                "Lock %s expired but access denied (%lu < %u)",
+                                szLock, curTime - (time_t)(byNameStat.st_mtime), 
+                                LOCK_LIFE );
+                    }
+                    break;
+
+                case EEXIST:
+                    /* 
+                     * Check the lock life 
+                     */
+                    curTime = time(0);
+                    nRet = stat ( szLock, &byNameStat );
+                    if ( nRet == -1 ) {
                         logit ( fTrace, POP_PRIORITY, HERE,
-                                "Open failed for %s: %s (%d)", 
-                                szLock, strerror(errno), errno );
-                        nRet = L_TMPLOCK;
+                               "fstat() for %s returned %s (%d) and "
+                               "open() returned EEXIST",
+                               szLock, strerror(errno), errno );
+                        attemptcnt = retrycnt;
+                        nRet = L_ERROR;
+                    } 
+                    else if ( curTime - (time_t)(byNameStat.st_mtime) > LOCK_LIFE ) {
+                        /* 
+                         * Create the lock file.
+                         *
+                         * This create could be happening to a link. 
+                         * Check the race 
+                         */
+                        if ( DEBUGGING && bDebugging ) { 
+                           logit ( fTrace, POP_DEBUG, HERE,
+                                   "Lock expired -- usurping; curTime=%u; "
+                                   "st_mtime=%u; age=%u; lock_life=%d",
+                                   (unsigned) curTime,
+                                   (unsigned) byNameStat.st_mtime,
+                                   (unsigned) curTime - (time_t)(byNameStat.st_mtime),
+                                   LOCK_LIFE );
+                        }
+                        fd = open ( szLock, O_WRONLY|O_CREAT|O_TRUNC, 0600 );
+                        if ( fd == -1 ) {
+                            logit ( fTrace, POP_PRIORITY, HERE,
+                                    "Open failed for %s: %s (%d)", 
+                                    szLock, strerror(errno), errno );
+                            nRet = L_TMPLOCK;
+                        }
+                        attemptcnt = retrycnt;
+                    } /* lock not owned by current process */
+                    else {
+                        if ( DEBUGGING && bDebugging ) { 
+                           logit ( fTrace, POP_DEBUG, HERE,
+                                   "lock %s busy (attempt %d of %d)", 
+                                   szLock, attemptcnt+1, retrycnt );
+                        }
+                        attemptcnt++;
+                        if ( attemptcnt == retrycnt ) 
+                            nRet = L_MAXTRYS;
+                        else
+                            nRet = L_ERROR;
                     }
+                    break;
+
+                default:
+                    logit ( fTrace, POP_PRIORITY, HERE,
+                            "Unexpected error opening lock file %s: %s (%d)", 
+                            szLock, strerror(errno), errno );
                     attemptcnt = retrycnt;
-                }
-                else {
-                    if ( DEBUGGING && bDebugging ) { 
-                       logit ( fTrace, POP_DEBUG, HERE,
-                               "lock %s busy (attempt %d of %d)", 
-                               szLock, attemptcnt+1, retrycnt );
+                    nRet = L_TMPLOCK;
+                    break;
+                } /* switch ( errno ) */
+
+            } /* open ( szLock ... ) == -1 */
+            else {
+                nRet = L_SUCCESS;
+                if ( DEBUGGING && bDebugging ) {
+                    logit ( fTrace, POP_DEBUG, HERE,
+                            "successfully opened (exclusive) lock %s", szLock );
                     }
-                    attemptcnt++;
-                    if ( attemptcnt == retrycnt ) 
-                        nRet = L_MAXTRYS;
-                }
                 break;
-
-            default:
-                logit ( fTrace, POP_PRIORITY, HERE,
-                        "Unexpected error opening lock file %s: %s (%d)", 
-                        szLock, strerror(errno), errno );
-                attemptcnt = retrycnt;
-                nRet = L_TMPLOCK;
-                break;
-            } /* switch ( errno ) */
-
-        } /* open ( szLock ... ) == -1 */
-        else {
-            attemptcnt = retrycnt;
-            if ( DEBUGGING && bDebugging ) {
-                logit ( fTrace, POP_DEBUG, HERE,
-                        "successfully opened (exclusive) lock %s", szLock );
-                }
-            break;
-        }
-#endif /* if defined(NO_ATOMIC_OPEN) */
-
-    } /* while ( attemptcnt != retrycnt ) */
+            }
+        } /* bNo_atomic_open == FALSE */
+    } /* while ( nRet != L_SUCCESS && attemptcnt != retrycnt ) */
 
     if ( fd != -1 ) { /* .lock file created successfully */
         /* 
@@ -469,7 +601,7 @@ Qmaillock ( char         *usernm,
             char buf [ BUF_SIZE ];
             int  nSize;
             
-            nSize = Qsprintf ( buf, "%d@", getpid() );
+            nSize = Qsprintf ( buf, "%d@", (int) getpid() );
             gethostname ( buf + nSize, BUF_SIZE - nSize - 1 );
             buf [ BUF_SIZE -1 ] = '\0';
             nSize = strlen(buf);
@@ -498,8 +630,8 @@ Qmaillock ( char         *usernm,
 
     if ( DEBUGGING && bDebugging ) {
         logit ( fTrace, POP_DEBUG, HERE,
-                "maillock() for user %s (%s) [%s:%lu] returning %d (%d attempt(s))", 
-                 usernm, szLock, fn, (unsigned long) ln, nRet, attemptcnt );
+                "maillock() on file %s (%s) [%s:%lu] returning %d (%d attempt(s))", 
+                 drop_name, szLock, fn, (unsigned long) ln, nRet, attemptcnt + 1 );
     }
     return nRet;
 }
