@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1998 Qualcomm Incorporated.
+ * Copyright (c) 2005 QUALCOMM Incorporated.
  * All rights reserved. The License.txt specifies the License terms
  * QPOPPER.
  *
@@ -10,7 +10,7 @@
  */
 #ifndef _CHARMANGLE_H_
 #define _CHARMANGLE_H_
-#include <config.h>
+#include "config.h"
 #include <mime.h>
 typedef enum { us_ascii=0,  iso_8859_1,  iso_8859_2,  iso_8859_3, 
 	       iso_8859_4,  iso_8859_5,  iso_8859_6,  iso_8859_7,
