@@ -3,7 +3,7 @@
 
      Laurence Lundblade <lgl@qualcomm.com>
 
-     Copyright 1997, QUALCOMM Inc.
+     Copyright (c) 2005 QUALCOMM Incorporated.
 
      File: manglers.h - handlers for individual content types
      Version: 0.2.3, Dec 1997
@@ -12,7 +12,7 @@
   ---- */
 #ifndef _MANGLERSINCLUDED
 #define _MANGLERSINCLUDED
-#include <config.h>
+#include "config.h"
 #include <mime.h>
 #include "charmangle.h"
 
