@@ -1,3 +1,10 @@
+/*
+ * Copyright (c) 2004 QUALCOMM Incorporated.  All rights reserved.
+ * The file License.txt specifies the terms for use, modification,
+ * and redistribution.
+ */
+
+
 /* ----------------------------------------------------------------------
      MIME Mangler -- single pass reduction of MIME to plain text
 
@@ -5,13 +12,7 @@
      Laurence Lundblade <lgl@qualcomm.com>,
      Randy Gellens
 
-     Copyright 1997,1998,1999, 2000 QUALCOMM Incorporated.  All rights reserved.
-     The file License.txt specifies the terms for use, modification, and 
-     redistribution.
-
      File: striphtml.c
-     Version: 0.3.1, May 1999
-     Last Edited: 14 March 2000
 
   ---- */
      
@@ -46,14 +47,15 @@
 ||
 */
 
-#ifndef UNIX
-#include <Pilot.h>
+#ifndef  UNIX
+#  include <Pilot.h>
 #else /* UNIX */
-#include <stdio.h>
-#define StrCompare strcmp
-#define StrCopy    strcpy
-#define StrCat     strcat
-#define StrLen     strlen
+#  include <stdio.h>
+#  include <string.h>
+#  define StrCompare strcmp
+#  define StrCopy    strcpy
+#  define StrCat     strcat
+#  define StrLen     strlen
 
 #endif /* UNIX */
 
@@ -62,8 +64,6 @@
 #include "lineend.h"
 
 #ifdef _DEBUG
-#  include <stdio.h>
-#  include <string.h>
 #  include "popper.h"
 #endif
 
@@ -604,7 +604,7 @@ void ffToHTML_lineStart(struct Flowed2HTML *me)
              
      int n;
      for (n = me->m_nQuoteDepth - me->m_nLastQuoteDepth; n; n--)
-       OUTBUF("<blockquote>", 12);
+       OUTBUF("<blockquote type=cite>", 12);
    }
            
    else 
