@@ -6,12 +6,70 @@
 
 
 /*
- * Copyright (c) 2000 Qualcomm Incorporated.  All rights reserved.
+ * Copyright (c) 2005 QUALCOMM Incorporated.  All rights reserved.
  * The file License.txt specifies the terms for use, modification,
  * and redistribution.
  *
  * Revisions:
  *
+ *     04/22/05  [rcg]
+ *              - Use generic auth error msg and sleep in more cases
+ *
+ *     04/18/05  [rcg]
+ *              - Applied revised patch by David Shrimpton to use
+ *                pam_setcred() instead of pam_*_session() to deal
+ *                with Kerberos tickets.  Note that, while it would
+ *                be good for us to be able to call pam_*_session(),
+ *                we can't do so until we have a way to call
+ *                pam_close_session() when we're really done, which we
+ *                currently can't do because we've given up privileges
+ *                by then.
+ *
+ *     12/04/04  [rcg]
+ *              - Applied patch by David Shrimpton at its.uq.edu.au
+ *                to pam authentication code to ensure pam_close_session
+ *                is called to destroy kerberos tickets.   Otherwise dead
+ *                tickets accumulate in cache directory which runs out of
+ *                space quickly on busy server.  Problem noted and fix
+ *                suggested by Rodney McDuff ITS UQ.   (Directory
+ *                permissions on ticket cache dir need to be 1777)
+ *
+ *     07/23/04  [rcg]
+ *              - when configured as --with-pam and building for 
+ *                Darwin, include <pam/pam_appl.h> instead of 
+ *                <security/pam_appl.h> (otherwise build fails)
+ *                (thanks to Kyle McKay for the patch)
+ *
+ *     01/16/03  [rcg]
+ *             - Renamed PASSWD macro to QPASSWD to avoid redefining
+ *               PASSWD in shadow.h.
+ *
+ *     06/01/01  [rcg]
+ *             - Added patch by Clifton Royston to change error message
+ *               for nonauthfile and authfile tests.
+ *
+ *     02/03/01  [rcg]
+ *             - Now logging to p->pop_facility.
+ *
+ *     01/30/01  [rcg]
+ *             - Now using p->bDo_timing instead of DO_TIMING.
+ *             - Now using p->bCheck_pw_max instead of DONT_CHECK_SP_MAX.
+ *             - Now using p->grp_serv_mode and p->grp_no_serv_mode instead
+ *               of SERVER_MODE_GROUP_INCLUDE and SERVER_MODE_GROUP_EXCLUDE.
+ *             - Now using p->authfile instead of AUTHFILE.
+ *
+ *     12/27/00  [rcg]
+ *             - Added generic auth_user() to pop_pass.c.
+ *
+ *     11/12/00  [rcg]
+ *             - Fixed reference to 'pname' instead of 'p->myname'.
+ *
+ *     10/03/00  [rcg]
+ *             - Reopen log after calling pam_authenticate().
+ *
+ *     10/02/00  [rcg]
+ *             - Call check_config_files to check qpopper-options files.
+ *
  *     09/26/00  [rcg]
  *             - Issue error 4 instead of crashing if appdata pointer
  *               is null (known bug in Solaris PAM fixed in Solaris 7
@@ -94,10 +152,7 @@
 #  include <sys/unistd.h>
 #endif
 
-#ifdef DO_TIMING
-#  include <time.h>
-#endif /* DO_TIMING */
-
+#include <time.h>
 #include <pwd.h>
 #include "popper.h"
 
@@ -113,15 +168,8 @@ extern char            *client_name;
 
 #include "snprintf.h"
 
-#ifdef DONT_CHECK_SP_MAX
-#  define   CHECK_SP_MAX    FALSE
-#else
-#  define   CHECK_SP_MAX    TRUE
-#endif /* DONT_CHECK_SP_MAX */
-
 #define SLEEP_SECONDS 10
 
-
 /* This error message is vague on purpose to help improve
    security at the inconvience of administrators and users */
 
@@ -130,57 +178,104 @@ char    *ERRMSG_ACEXP = "[AUTH] \"%s\": account expired.";
 char    *ERRMSG_PWEXP = "[AUTH] \"%s\": password expired.";
 char    *ERRMSG_AUTH  = "[AUTH] \"%s\": access denied.";
 
+/*------------------------------------- PASSWORD */
+pop_result
+check_password ( POP *p, char *password_entry, char *password_claim )
+{
+    /*
+     * First check if password entry (system password) is null
+     */
+    if ( (  password_entry == NULL ) || 
+         ( *password_entry == '\0' )   ) {
+        sleep ( SLEEP_SECONDS );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+    }
+
+    /*
+     * Now compare the claimed password with the system password entry
+     *
+     * Linux kernels older than 2.0.x need pw_encrypt() for shadow support 
+     */
+    if ( TRUE
+#ifdef    HAVE_CRYPT16
+         && strcmp ( crypt16    ( password_claim, password_entry ), password_entry )
+#endif /* HAVE_CRYPT16 */
+#ifdef    HAVE_BIGCRYPT
+         && strcmp ( bigcrypt   ( password_claim, password_entry ), password_entry )
+#endif /* HAVE_BIGCRYPT */
+#ifdef   HAVE_PW_ENCRYPT
+         && strcmp ( pw_encrypt ( password_claim, password_entry ), password_entry )
+#endif /* HAVE_PW_ENCRYPT */
+#ifdef    HAVE_CRYPT
+         && strcmp ( crypt      ( password_claim, password_entry ), password_entry )
+#endif /* HAVE_CRYPT */
+       ) {
+        sleep  ( SLEEP_SECONDS );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+    }
+
+    return POP_SUCCESS;
+}
+
+
+
 
 /*------------------------------------- NONAUTHFILE */
-#ifdef NONAUTHFILE
 int
-checknonauthfile ( user ) 
-     char *user;
+checknonauthfile ( POP *p ) 
 {
     char buf [ MAXUSERNAMELEN +1 ];
     FILE *fp;
 
-    fp = fopen ( NONAUTHFILE, "r" );
+    fp = fopen ( p->nonauthfile, "r" );
     if ( fp != NULL ) {
         while ( fgets ( buf, MAXUSERNAMELEN +1, fp ) ) {
             buf [ strlen(buf) -1 ] = '\0';
-            if ( !strcmp ( buf, user ) ) {
+            if ( !strcmp ( buf, p->user ) ) {
                 fclose ( fp );
+                DEBUG_LOG3 ( p, "...checknonauthfile matched user %s "
+                                "in file %s: %s",
+                             p->user, p->nonauthfile, buf );
                 return ( -1 );
             }
         }
 
         fclose ( fp );
     }
+
+    DEBUG_LOG2 ( p, "...checknonauthfile didn't match user %s in file %s",
+                 p->user, p->nonauthfile );
     return ( 0 );
 }
-#endif /* NONAUTHFILE */
 
 
 /*------------------------------------- AUTHFILE */
-#ifdef AUTHFILE
 int
-checkauthfile ( user ) 
-char *user;
+checkauthfile ( POP *p ) 
 {
     char buf [ MAXUSERNAMELEN +1 ];
     FILE *fp;
 
-    fp = fopen ( AUTHFILE, "r" );
+    fp = fopen ( p->authfile, "r" );
     if ( fp != NULL ) {
         while ( fgets ( buf, MAXUSERNAMELEN +1, fp ) ) {
             buf [ strlen(buf) -1 ] = '\0';
-            if ( !strcmp ( buf, user ) ) {
+            if ( !strcmp ( buf, p->user ) ) {
                 fclose ( fp );
+                DEBUG_LOG3 ( p, "...checkauthfile matched user %s "
+                                "in file %s: %s",
+                             p->user, p->authfile, buf );
                 return ( 0 );
             }
         }
 
         fclose ( fp );
     }
+
+    DEBUG_LOG2 ( p, "...checkauthfile didn't match user %s in file %s",
+                 p->user, p->authfile );
     return ( -1 );
 }
-#endif /* AUTHFILE */
 
 
 int 
@@ -212,6 +307,7 @@ struct passwd *pw;
         pop_log ( p, POP_WARNING, HERE, "%s: (%s.%s@%s) %s", 
                   p->client, kdata.pname, 
                   kdata.pinst, kdata.prealm, krb_err_txt[status] );
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE,
                 "[AUTH] Kerberos error:  \"%s\".", krb_err_txt[status] ) );
     } /* krb_get_lrealm failed */
@@ -223,6 +319,7 @@ struct passwd *pw;
                   "%s: (%s.%s@%s): not in %s's ACL.",
                   p->client, kdata.pname, kdata.pinst, 
                   kdata.prealm, p->user );
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, "[AUTH] Not in %s's ACL.", 
                            p->user ) );
     }
@@ -274,7 +371,8 @@ struct passwd *pw;
     if ( !krb5_kuserok ( pop_context, ext_client, p->user ) ) {
         pop_log ( p, POP_WARNING, HERE, "%s: (%s): not in %s's ACL.",
                   p->client, client_name, p->user );
-        return ( pop_msg ( p,POP_FAILURE, HERE, "Not in %s's ACL.", p->user ) );
+        sleep ( SLEEP_SECONDS );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
     }
 
 #   else /* KUSEROK */
@@ -287,6 +385,7 @@ struct passwd *pw;
             pop_log ( p, POP_WARNING, HERE, "%s: (%s) %s", 
                       p->client, client_name,
                       error_message(retval) );
+            sleep ( SLEEP_SECONDS );
             return ( pop_msg ( p,POP_FAILURE, HERE,
                                "Kerberos error:  \"%s\".", 
                                error_message(retval) ) );
@@ -296,7 +395,7 @@ struct passwd *pw;
         if ( strncmp ( tmpdata->data, lrealm, tmpdata->length ) )  {
             pop_log ( p, POP_WARNING, HERE, "%s: (%s) realm not accepted.",
                       p->client, client_name );
-            return ( pop_msg ( p,POP_FAILURE, HERE,
+            return ( pop_msg ( p, POP_FAILURE, HERE,
                                "Kerberos realm \"%*s\" not accepted.",
                                tmpdata->length, tmpdata->data ) );
         }
@@ -327,17 +426,19 @@ struct passwd *pw;
 
 /*------------------------------------- SPEC_POP_AUTH */
 #ifdef SPEC_POP_AUTH
-    char *crypt();
 
 
 /*----------------------------------------------- PAM */
 /* Code based on PAM patch contributed by German Poo.  
  * Patched by Kenneth Porter.
+ * Patched by others since.
  */
 #  ifdef USE_PAM
 
 #    ifdef HAVE_SECURITY_PAM_APPL_H
 #      include <security/pam_appl.h>
+#    elif defined HAVE_PAM_PAM_APPL_H
+#      include <pam/pam_appl.h>
 #    endif /* HAVE_SECURITY_PAM_APPL_H */
 
 /* 
@@ -349,7 +450,7 @@ static char *GP_ERRSTRING =
 
 static int 
 PAM_qpopper_conv ( int                        num_msg, 
-                   struct pam_message       **msg, 
+                   const struct pam_message **msg, 
                    struct pam_response      **resp, 
                    void                      *appdata_ptr )
 {
@@ -418,6 +519,8 @@ struct passwd *pw;
     int              erc            = 0;
     const char      *errmsg         = NULL;
 
+    (void) pw; /* avoid 'unused parameter' warnings */
+
     /* 
      * Let conv function access POP structure 
      */
@@ -433,7 +536,24 @@ struct passwd *pw;
         return  ( pop_msg ( p, POP_FAILURE, HERE, GP_ERRSTRING, p->user,
                             pam_strerror(NULL, erc), erc ) );
     }
+    pamerror = pam_set_item ( pamh, PAM_TTY, p->ipaddr );
+    DEBUG_LOG1 ( p, "pam_set_item returned %i", pamerror );
+    if ( pamerror != PAM_SUCCESS ) {
+        pop_log ( p, POP_NOTICE, HERE,
+                  "Error setting pam item TTY: %s (%i)", 
+                  pam_strerror ( pamh, pamerror ), pamerror );
+    }
+
+    pamerror = pam_set_item ( pamh, PAM_RHOST, p->client );
+    DEBUG_LOG1 ( p, "pam_set_item returned %i", pamerror );
+    if ( pamerror != PAM_SUCCESS ) {
+        pop_log ( p, POP_NOTICE, HERE,
+                  "Error setting pam item RHOST: %s (%i)", 
+                  pam_strerror ( pamh, pamerror ), pamerror );
+    }
     pamerror = pam_authenticate ( pamh, 0 );
+
+
     DEBUG_LOG2 ( p, "pam_authenticate returned %i; gp_errcode=%i", 
                  pamerror, gp_errcode );
     if ( ( gp_errcode != 0 ) || ( pamerror != PAM_SUCCESS ) ) {
@@ -444,30 +564,39 @@ struct passwd *pw;
         return  ( pop_msg ( p, POP_FAILURE, HERE, GP_ERRSTRING, 
                             p->user, errmsg, erc ) );
     }
-    pamerror = pam_acct_mgmt ( pamh, 0 );
-    DEBUG_LOG2 ( p, "pam_acct_mgmt returned %i; gp_errcode=%i", 
-                 pamerror, gp_errcode );
-    if ( ( gp_errcode != 0 ) || ( pamerror != PAM_SUCCESS ) ) {
-        erc = pamerror ? pamerror : gp_errcode;
-        errmsg = pam_strerror ( pamh, erc );
-        pam_end ( pamh, 0 );
-        sleep   ( SLEEP_SECONDS );
-        return  ( pop_msg ( p, POP_FAILURE, HERE, GP_ERRSTRING, 
-                            p->user, errmsg, erc ) );
-    }
     pamerror = pam_setcred ( pamh, PAM_ESTABLISH_CRED );
-    DEBUG_LOG2 ( p, "pam_setcred returned %i; gp_errcode=%i", 
-                 pamerror, gp_errcode );
-    if ( ( gp_errcode != 0 ) || ( pamerror != PAM_SUCCESS ) ) {
-        erc = pamerror ? pamerror : gp_errcode;
-        errmsg = pam_strerror ( pamh, erc );
+    DEBUG_LOG1 ( p, "pam_setcred returned %i", pamerror );
+    if ( pamerror != PAM_SUCCESS ) {
+        errmsg = pam_strerror ( pamh, pamerror );
         pam_end ( pamh, 0 );
         sleep   ( SLEEP_SECONDS );
         return  ( pop_msg ( p, POP_FAILURE, HERE, GP_ERRSTRING, 
-                            p->user, errmsg, erc ) );
+                            p->user, errmsg, pamerror ) );
+    }
+    pamerror = pam_acct_mgmt ( pamh, 0 );
+    /*
+     * Need to reopen log after some pam calls, e.g., pam_acct_mgmt() which 
+     * open log and set facility to LOG_AUTH and then closelog.
+     * They may not log anything.
+     */
+#ifdef SYSLOG42
+    openlog ( p->myname, 0 );
+#else
+    openlog ( p->myname, POP_LOGOPTS, p->log_facility );
+#endif
+    DEBUG_LOG1 ( p, "pam_acct_mgmt returned %i", pamerror );
+    if ( pamerror != PAM_SUCCESS ) {
+        errmsg = pam_strerror ( pamh, pamerror );
+        pam_end ( pamh, 0 );
+        sleep   ( SLEEP_SECONDS );
+        return  ( pop_msg ( p, POP_FAILURE, HERE, GP_ERRSTRING, 
+                            p->user, errmsg, pamerror ) );
     }
-    pam_end ( pamh, PAM_SUCCESS );
     
+    pamerror = pam_setcred ( pamh, PAM_DELETE_CRED );
+    DEBUG_LOG1 ( p, "pam_setcred returned %i", pamerror );
+    pam_end ( pamh, PAM_SUCCESS );
+
     return POP_SUCCESS;
 }
 #  else /* not PAM */
@@ -476,9 +605,11 @@ struct passwd *pw;
 /*----------------------------------------------- SUNOS4 and not ISC */
 #    if defined(SUNOS4) && !defined(ISC)
 
-#    include <sys/label.h>
-#    include <sys/audit.h>
-#    include <pwdadj.h>
+#      define DECLARED_AUTH_USER
+
+#      include <sys/label.h>
+#      include <sys/audit.h>
+#      include <pwdadj.h>
 
 static int
 auth_user ( p, pw )
@@ -492,28 +623,16 @@ struct passwd *pw;
      */
     pwadj = getpwanam ( p->user );
     if ( pwadj == NULL ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE,
-                "[AUTH] (shadow) Password supplied for \"%s\" is empty.", 
-                p->user ) );
+        pop_log ( p, POP_WARNING, HERE,
+                "(shadow) Password supplied for \"%s\" is empty.", 
+                p->user );
+        sleep ( SLEEP_SECONDS );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
     } else {
         pw->pw_passwd = (char *)strdup(pwadj->pwa_passwd);
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) ||
-           strcmp ( crypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) 
-       ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* SUNOS4 and not ISC */
@@ -522,6 +641,8 @@ struct passwd *pw;
 /*----------------------------------------------- SOLARIS2 or AUX or UXPDS or SGI */
 #    if defined(SOLARIS2) || defined(AUX) || defined(UXPDS) || defined(sgi)
 
+#      define DECLARED_AUTH_USER
+
 #      ifdef HAVE_SHADOW_H
 #        include <shadow.h>
 #      endif /* HAVE_SHADOW_H */
@@ -582,8 +703,9 @@ struct passwd *pw;
         /* 
          * Check if password not changed within required time
          */
-        if ( CHECK_SP_MAX && 
-             pwd->sp_max > 0 && today > pwd->sp_lstchg + pwd->sp_max ) {
+        if ( p->bCheck_pw_max
+             && pwd->sp_max > 0
+             && today > pwd->sp_lstchg + pwd->sp_max ) {
             sleep  ( SLEEP_SECONDS );
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PWEXP, p->user ) );
         }
@@ -592,20 +714,7 @@ struct passwd *pw;
         endspent();
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) ||
-           strcmp ( crypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) 
-       ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* SOLARIS2 or AUX or UXPDS or sgi */
@@ -614,6 +723,8 @@ struct passwd *pw;
 /*----------------------------------------------- PIX or ISC */
 #    if defined(PTX) || defined(ISC) 
 
+#      define DECLARED_AUTH_USER
+
 #      ifdef HAVE_SHADOW_H
 #        include <shadow.h>
 #      endif /* HAVE_SHADOW_H */
@@ -639,20 +750,7 @@ struct passwd *pw;
         pw->pw_passwd = (char *)strdup(pwd->sp_pwdp);
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) ||
-           strcmp ( crypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) 
-       ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* PTX or ISC */
@@ -661,6 +759,8 @@ struct passwd *pw;
 /*----------------------------------------------- POPSCO or HPUX */
 #    if defined(POPSCO) || defined(HPUX)
 
+#      define DECLARED_AUTH_USER
+
 #      ifdef POPSCO
 #        include <sys/security.h>
 #        include <sys/audit.h>
@@ -670,7 +770,8 @@ struct passwd *pw;
 
 #    include <prot.h>
 
-#    define PASSWD(p)       p->ufld.fd_encrypt
+#    define QPASSWD(p)      p->ufld.fd_encrypt
+
 
 static int
 auth_user ( p, pw )
@@ -680,43 +781,17 @@ struct passwd *pw;
     register struct pr_passwd *pr;
 
     pr = getprpwnam ( p->user );
-    if ( pr == NULL ) {
-        if ( !strcmp ( pw->pw_passwd, "x" ) ) {
+    if ( pr == NULL ) { /* Can't get shadow password */
+        if ( !strcmp ( pw->pw_passwd, "x" ) ) { /* user has shadow password */
             sleep  ( SLEEP_SECONDS );
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
         }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) ||
-            ( strcmp ( bigcrypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) &&
-              strcmp ( crypt    ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd )
-            ) 
-        ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-        }
-    } else {
-        /* 
-         * We don't accept connections from users with null passwords 
-         *
-         * Compare the supplied password with the password file entry 
-         */
-        if ( ( PASSWD(pr) == NULL ) || ( *PASSWD(pr) == '\0' ) ) {
-            sleep  ( SLEEP_SECONDS );
-            return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-        }
-        
-        if ( strcmp ( bigcrypt ( p->pop_parm[1], PASSWD(pr) ), PASSWD(pr) ) &&
-             strcmp ( crypt    ( p->pop_parm[1], PASSWD(pr) ), PASSWD(pr) )  ) {
-            sleep  ( SLEEP_SECONDS );
-            return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-        }
+        return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
+    }
+    else { /* got shadow password */
+        return ( check_password ( p, QPASSWD(pr), p->pop_parm[1] ) );
     }
-
-    return ( POP_SUCCESS );
 }
 
 #    endif  /* POPSCO || HPUX */
@@ -725,12 +800,12 @@ struct passwd *pw;
 /*----------------------------------------------- ULTRIX */
 #    ifdef ULTRIX
 
+#      define DECLARED_AUTH_USER
+
 #      include <auth.h>
 
 static int
-auth_user ( p, pw )
-struct passwd  *   pw;
-POP     *   p;
+auth_user ( POP *p, struct passwd *pw )
 {
     AUTHORIZATION *auth, *getauthuid();
 
@@ -741,27 +816,10 @@ POP     *   p;
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
         }
     } else {
-        pw->pw_passwd = (char *)strdup(auth->a_password);
+        pw->pw_passwd = (char *) strdup ( auth->a_password );
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' )  {
-        sleep ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    if ( strcmp ( crypt16 ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) &&
-         strcmp ( crypt   ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd )  ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* ULTRIX */
@@ -771,16 +829,16 @@ POP     *   p;
 /* this is a DEC Alpha / Digital Unix system */
 #    ifdef OSF1
 
+#      define DECLARED_AUTH_USER
+
 #      include <sys/types.h>
 #      include <sys/security.h>
 #      include <prot.h>
 
-#      define   PASSWD(p)   (p->ufld.fd_encrypt)
+#      define   QPASSWD(p)   (p->ufld.fd_encrypt)
 
 static int
-auth_user ( p, pw )
-POP     *   p;
-struct passwd *pw;
+auth_user ( POP *p, struct passwd *pw )
 {
     register struct pr_passwd *pr;
 
@@ -791,27 +849,10 @@ struct passwd *pw;
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
         }
     } else {
-        pw->pw_passwd = (char *)strdup(PASSWD(pr));
+        pw->pw_passwd = (char *) strdup ( QPASSWD(pr) );
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' )  ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    if ( strcmp ( bigcrypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) &&
-         strcmp ( crypt    ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd )  ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif        /* OSF1 */
@@ -820,12 +861,14 @@ struct passwd *pw;
 /*----------------------------------------------- UNIXWARE */
 #    ifdef UNIXWARE
 
-#      include <shadow.h>
+#      define DECLARED_AUTH_USER
+
+#      ifdef HAVE_SHADOW_H
+#        include <shadow.h>
+#      endif /* HAVE_SHADOW_H */
 
 static int
-auth_user ( p, pw )
-struct passwd  *   pw;
-POP     *   p;
+auth_user ( POP *p, struct passwd *pw )
 {
     register struct spwd * pwd;
     long today;
@@ -853,8 +896,9 @@ POP     *   p;
         /* 
          * Check if password not changed within required time 
          */
-        if ( CHECK_SP_MAX && 
-             pwd->sp_max > 0 && today > pwd->sp_lstchg + pwd->sp_max ) {
+        if ( p->bCheck_pw_max
+             && pwd->sp_max > 0
+             && today > pwd->sp_lstchg + pwd->sp_max ) {
             sleep  ( SLEEP_SECONDS );
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PWEXP, p->user ) );
         }
@@ -863,20 +907,7 @@ POP     *   p;
         endspent();
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) ||
-         strcmp ( crypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) 
-       ) {
-        sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* UNIXWARE */
@@ -885,15 +916,14 @@ POP     *   p;
 /*----------------------------------------------- LINUX */
 #    if defined(LINUX)
 
+#      define DECLARED_AUTH_USER
 
 #      ifdef HAVE_SHADOW_H
 #        include <shadow.h>
 #      endif /* HAVE_SHADOW_H */
 
 static int
-auth_user ( p, pw )
-POP     *   p;
-struct passwd *pw;
+auth_user ( POP *p, struct passwd *pw )
 {
     register struct spwd * pwd;
     long today;
@@ -908,7 +938,7 @@ struct passwd *pw;
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
         }
     } else {
-        today = (long)time((time_t *)NULL)/24/60/60;
+        today = (long) time ( (time_t *) NULL ) / 24 / 60 / 60;
 
         /* 
          * Check for expiration date 
@@ -921,38 +951,18 @@ struct passwd *pw;
         /* 
          * Check if password not changed within required time 
          */
-        if ( CHECK_SP_MAX && 
-             pwd->sp_max > 0 && today > pwd->sp_lstchg + pwd->sp_max ) {
+        if ( p->bCheck_pw_max
+             && pwd->sp_max > 0
+             && today > pwd->sp_lstchg + pwd->sp_max ) {
             sleep  ( SLEEP_SECONDS );
             return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PWEXP, p->user ) );
         }
 
-        pw->pw_passwd = (char *)strdup(pwd->sp_pwdp);
+        pw->pw_passwd = (char *) strdup ( pwd->sp_pwdp );
         endspent();
     }
 
-    /*  
-     * We don't accept connections from users with null passwords 
-     *
-     *  Compare the supplied password with the password file entry
-     *
-     *  Kernels older than 2.0.x need pw_encrypt() for shadow support 
-     */
-    if ( (  pw->pw_passwd == NULL ) || 
-         ( *pw->pw_passwd == '\0' ) || 
-         ( strcmp ( crypt      ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd ) &&
-#      ifdef HAVE_PW_ENCRYPT
-           strcmp ( pw_encrypt ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd )
-#      else /* not HAVE_PW_ENCRYPT */
-           strcmp ( crypt      ( p->pop_parm[1], pw->pw_passwd ), pw->pw_passwd )
-#      endif /* HAVE_PW_ENCRYPT */
-         )
-       ) {
-        sleep  ( SLEEP_SECONDS);
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
-    }
-
-    return ( POP_SUCCESS );
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
 }
 
 #    endif  /* LINUX */
@@ -961,18 +971,20 @@ struct passwd *pw;
 /*----------------------------------------------- BSD/OS */
 #    if defined(__bsdi__) && _BSDI_VERSION >= 199608
 
+#      define DECLARED_AUTH_USER
+
 static int
-auth_user(p, pw)
-POP     *   p;
-struct passwd  *   pw;
+auth_user ( POP *p, struct passwd *pw)
 {
     int r;
 
     /*
      * We don't accept connections from users with null passwords
      */
-    if ( ( pw->pw_passwd == NULL ) || ( *pw->pw_passwd == '\0' ) )
+    if ( ( pw->pw_passwd == NULL ) || ( *pw->pw_passwd == '\0' ) ) {
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+    }
 
     /*
      * Compare the supplied password with the password file entry
@@ -985,8 +997,10 @@ struct passwd  *   pw;
             p->challenge = NULL;
     }
 
-    if ( r <= 0 || auth_approve ( p->class, p->user, "popper" ) <= 0 )
+    if ( r <= 0 || auth_approve ( p->class, p->user, "popper" ) <= 0 ) {
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+    }
 
     return ( POP_SUCCESS );
 }
@@ -997,6 +1011,8 @@ struct passwd  *   pw;
 /*----------------------------------------------- AIX  */
 #    ifdef AIX
 
+#      define DECLARED_AUTH_USER
+
 /*
  * AIX specific special authentication
  *
@@ -1028,6 +1044,7 @@ auth_user ( POP *p, struct passwd *pw )
         int save_err = errno;
         pop_log ( p, POP_NOTICE, HERE, "Account %s access denied: (%d) %s",
                   p->user, save_err, message );
+        sleep ( SLEEP_SECONDS );
         if ( save_err == ESTALE )
             pop_msg ( p, POP_FAILURE, HERE, ERRMSG_ACEXP, p->user );
         else
@@ -1069,18 +1086,47 @@ Exit:
 #    endif /* AIX */
 
 
+/*----------------------------------------------- generic AUTH_USER */
+#    ifndef DECLARED_AUTH_USER 
+
+#      ifdef HAVE_SHADOW_H
+#        include <shadow.h>
+#      endif /* HAVE_SHADOW_H */
+
+static int
+auth_user ( POP *p, struct passwd *pw )
+{
+    register struct spwd *pwd;
+    long today;
+
+    /*  
+     * Look for the user in the shadow password file 
+     */
+    pwd = getspnam ( p->user );
+    if ( pwd == NULL ) {
+        if ( !strcmp ( pw->pw_passwd, "x" ) ) {      /* This my be a YP entry */
+            sleep  ( SLEEP_SECONDS );
+            return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+        }
+    } else {
+        pw->pw_passwd = (char *)strdup(pwd->sp_pwdp);
+    }
+
+    return ( check_password ( p, pw->pw_passwd, p->pop_parm[1] ) );
+}
+
+#    endif  /* DECLARED_AUTH_USER */
+
+
 #  endif /* USE_PAM */
 
 
 /*------------------------------------- not SPEC_POP_AUTH */
 #else   /* not SPEC_POP_AUTH */
 
-char *crypt();
 
 static int
-auth_user ( p, pw )
-POP     *   p;
-struct passwd  *   pw;
+auth_user ( POP *p, struct passwd *pw )
 {
     /*  
      * We don't accept connections from users with null passwords 
@@ -1110,15 +1156,11 @@ struct passwd  *   pw;
 #  include <rpc/key_prot.h>
 #endif /* SECURENISPLUS */
 
-int 
-pop_pass ( p )
-POP     *  p;
+pop_result 
+pop_pass ( POP *p )
 {
-    struct passwd pw, *pwp;
-
-#ifdef DO_TIMING
+    struct passwd *pwp;
     time_t my_timer = time(0);
-#endif /* DO_TIMING */
 
 #ifdef CHECK_SHELL
     char *getusershell();
@@ -1136,53 +1178,39 @@ POP     *  p;
     *secretkey = '\0';
 #endif /* SECURENISPLUS */
 
-#ifdef NONAUTHFILE
     /* 
      * Is the user not authorized to use POP? 
      */
-    if ( checknonauthfile ( p->user ) != 0 ) {
-        DEBUG_LOG2 ( p, "User %.128s in nonauthfile %.256s",
-                     p->user, NONAUTHFILE );
+    if ( p->nonauthfile != NULL && checknonauthfile ( p ) != 0 ) {
         sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_AUTH, p->user ) );
     }
-#endif /* NONAUTHFILE */
 
-#ifdef AUTHFILE
     /* 
      * Is the user authorized to use POP? 
      */
-    if ( checkauthfile ( p->user ) != 0 ) {
-        DEBUG_LOG2 ( p, "User %.128s not in authfile %.256s",
-                     p->user, AUTHFILE );
+    if ( p->authfile != NULL && checkauthfile ( p ) != 0 ) {
         sleep  ( SLEEP_SECONDS );
-        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_AUTH, p->user ) );
     }
-#endif /* AUTHFILE */
 
     /*  
-     * Look for the user in the password file 
+     * Verify user known by system. 
      */
-    pwp = getpwnam ( p->user );
-    if ( pwp == NULL ) {
-        DEBUG_LOG1 ( p, "User %.128s not in passwd file",
+    pwp = &p->pw;
+    if ( pwp->pw_name == NULL ) {
+        DEBUG_LOG1 ( p, "User %.128s not known by system",
                      p->user );
         sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
     }
 
-    pw = *pwp;
-    pw.pw_dir = strdup(pwp->pw_dir);
-    
-    DEBUG_LOG2 ( p, "home (%d): '%s'",
-                     strlen(pw.pw_dir), pw.pw_dir );
-
 #ifdef SECURENISPLUS
     /*  
      * we must do this keyserv stuff (as well as auth_user()!) as the user 
      */
     uid_save = geteuid();
-    seteuid ( pw.pw_uid );
+    seteuid ( pwp->pw_uid );
 
     /*  
      * see if DES keys are already known to the keyserv(1m) 
@@ -1201,7 +1229,6 @@ POP     *  p;
             if ( ! getsecretkey ( net_name, secretkey, p->pop_parm[1] ) ||
                  *secretkey == '\0' ) {
                 sleep  ( SLEEP_SECONDS );
-                free   ( pw.pw_dir );
                 return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
             }
 
@@ -1219,11 +1246,12 @@ POP     *  p;
 #endif /* SECURENISPLUS */
 
 #ifdef BLOCK_UID
-    if ( pw.pw_uid <= BLOCK_UID ) {
-        free ( pw.pw_dir );
-        return ( pop_msg ( p, POP_FAILURE, HERE,
-                           "[AUTH] Access is blocked for UIDs below %d", 
-                           BLOCK_UID ) );
+    if ( pwp->pw_uid <= BLOCK_UID ) {
+        pop_log ( p, POP_WARNING, HERE,
+                  "[AUTH] Access is blocked for UIDs below %d", 
+                  BLOCK_UID );
+        sleep ( SLEEP_SECONDS );
+        return ( pop_msg ( p, POP_FAILURE, HERE, ERRMSG_PW, p->user ) );
     }
 #endif /* BLOCK_UID */
 
@@ -1233,15 +1261,9 @@ POP     *  p;
      * getusershell(), unless the sys admin has included the wildcard
      * shell in /etc/shells.  (default wildcard - /POPPER/ANY/SHELL)
      */
-    shell = pw.pw_shell;
+    shell = pwp->pw_shell;
     if ( shell == NULL || *shell == 0 ) {
-        /* 
-         * You can default the shell, but I don't think it's a good idea 
-         */
-        /* 
-         shell = "/usr/bin/sh"; 
-         */
-        free ( pw.pw_dir );
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, 
                            "[AUTH] No user shell defined" ) );
     }
@@ -1255,22 +1277,26 @@ POP     *  p;
     endusershell();
 
     if ( shellvalid == 0 ) {
-        free ( pw.pw_dir );
+        sleep ( SLEEP_SECONDS );
         return ( pop_msg ( p, POP_FAILURE, HERE, 
                            "[AUTH] \"%s\": shell not found.", 
                            p->user ) );
     }
 #endif /* CHECK_SHELL */
 
-    if ( ( p->kerberos ? auth_user_kerberos ( p, pwp ) 
-                       : auth_user          ( p, pwp ) 
+    /*
+     * Authenticate the user.
+     *
+     * Invoke auth_user(), which is defined above based on specific
+     * system type and ./configure options
+     */
+    if ( ( p->bKerberos ? auth_user_kerberos ( p, pwp ) 
+                        : auth_user          ( p, pwp ) 
          ) != POP_SUCCESS 
        ) {
         pop_log ( p, POP_PRIORITY, HERE,
                  "[AUTH] Failed attempted login to %s from host (%s) %s",
                   p->user, p->client, p->ipaddr );
-        free   ( pw.pw_dir );
-        sleep  ( SLEEP_SECONDS );
         return ( POP_FAILURE );
     }
 
@@ -1282,61 +1308,36 @@ POP     *  p;
      * Check if server mode should be set or reset based on group membership.
      */
 
-#ifdef SERVER_MODE_GROUP_INCL
-    if ( check_group ( p, &pw, SERVER_MODE_GROUP_INCL ) ) {
+    if ( p->grp_serv_mode != NULL
+         && check_group ( p, &p->pw, p->grp_serv_mode ) ) {
         p->server_mode = TRUE;
-    DEBUG_LOG2 ( p, "Set server mode for user %s; member of \"%.128s\" group",
-                 p->user, SERVER_MODE_GROUP_INCL );
+        DEBUG_LOG2 ( p, "Set server mode for user %s; "
+                        "member of \"%.128s\" group",
+                     p->user, p->grp_serv_mode );
     }
-#endif /* SERVER_MODE_GROUP_INCL */
 
-#ifdef SERVER_MODE_GROUP_EXCL
-    if ( check_group ( p, &pw, SERVER_MODE_GROUP_EXCL ) ) {
+    if ( p->grp_no_serv_mode != NULL
+         && check_group ( p, &p->pw, p->grp_no_serv_mode ) ) {
         p->server_mode = FALSE;
-    DEBUG_LOG2 ( p, "Set server mode OFF for user %s; member of \"%.128s\" group",
-                 p->user, SERVER_MODE_GROUP_INCL );
+        DEBUG_LOG2 ( p, "Set server mode OFF for user %s; "
+                        "member of \"%.128s\" group",
+                     p->user, p->grp_no_serv_mode );
     }
-#endif /* SERVER_MODE_GROUP_EXCL */
 
     /*
-     * Process ~/.qpopper-options file, if requested and present.
+     * Process qpopper-options files, if requested and present.
      */
-    if ( p->user_opts ) {
-        int         rslt;
-        char        buf [ 256 ];
-        struct stat stat_buf;
+    check_config_files ( p, pwp );
 
-        rslt = Qsnprintf ( buf, sizeof(buf), "%s/.qpopper-options", pw.pw_dir );
-        if ( rslt == -1 )
-            pop_log ( p, POP_PRIORITY, HERE, 
-                      "Unable to build user options file name for user %s",
-                      p->user );
-        else {
-            rslt = stat ( buf, &stat_buf );
-            if ( rslt == 0 ) {
-                rslt = pop_config ( p, buf, TRUE );
-                if ( rslt == POP_FAILURE ) {
-                    pop_log ( p, POP_PRIORITY, HERE,
-                              "Unable to process user options file for user %s",
-                              p->user );
-                }
-            }
-        }
-    } /* p->user_opts */
-
-
-#ifdef DO_TIMING
-    p->login_time = time(0) - my_timer;
-#endif /* DO_TIMING */
+    if ( p->bDo_timing )
+        p->login_time = time(0) - my_timer;
 
     /*  
      * Make a temporary copy of the user's maildrop 
      * and set the group and user id 
      * and get information about the maildrop 
      */
-    if ( pop_dropcopy ( p, &pw ) != POP_SUCCESS ) {
-        free   ( pw.pw_dir );
-        sleep  ( SLEEP_SECONDS );
+    if ( pop_dropcopy ( p, pwp ) != POP_SUCCESS ) {
         return ( POP_FAILURE );
         }
 
@@ -1351,17 +1352,13 @@ POP     *  p;
      * Authorization completed successfully 
      */
 
-#ifdef LOG_LOGIN
-    pop_log ( p, POP_PRIORITY, HERE,
-              "(v%s) POP login by user \"%s\" at (%s) %s",
-              VERSION, p->user, p->client, p->ipaddr );
-#endif /* LOG_LOGIN */
+    if ( p->pLog_login != NULL )
+        do_log_login ( p );
 
 #ifdef DRAC_AUTH
     drac_it ( p );
 #endif /* DRAC_AUTH */
 
-    free ( pw.pw_dir );
     return ( pop_msg ( p, POP_SUCCESS, HERE,
                        "%s has %d visible message%s (%d hidden) in %ld octets.",
                         p->user,
